c     Copyright (C) 2008-2009 Vincent Eymet
c
c     Planet_EMC is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     Planet_EMC is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with Planet_EMC; if not, see <http://www.gnu.org/licenses/>
c
      subroutine distance_on_sphere(n,P1,P2,r,d)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the distance between two positions on a sphere
c
c     Inputs:
c       + n: dimension of vectors
c       + P1: (theta1,phi1) latitude and longitude of the first position
c       + P2: (theta2,phi2) latitude and longitude of the second position
c       + r: radius of the sphere
c     
c     Output:
c       + d: distance
c

c     inputs
      integer n
      double precision P1(1:2)
      double precision P2(1:2)
      double precision r
c     output
      double precision d
c     temp
      double precision pos1(1:Nmat_mx)
      double precision pos2(1:Nmat_mx)
      double precision alpha,prod
      integer stlren
      character*100 label
      label='subroutine distance_on_sphere'

      pos1(1)=r
      pos1(2)=P1(1)
      pos1(3)=P1(2)
      pos2(1)=r
      pos2(2)=P2(1)
      pos2(3)=P2(2)
      call scalar_product_spher(n,pos1,pos2,prod)
      alpha=dacos(prod) ! rad
      d=alpha*r ! m

      return
      end




      subroutine distance_spher(parent,r1,theta1,phi1,r2,theta2,phi2,d)
      implicit none
c
c     Purpose: to compute the distance between two points defined
c     by their spherical coordinates
c     
c     inputs
      double precision r1,theta1,phi1
      double precision r2,theta2,phi2
c     outputs
      double precision d
c     temp
      double precision x1,y1,z1
      double precision x2,y2,z2
      character*100 parent
      call spher2cart(parent,1,r1,theta1,phi1,x1,y1,z1)
      call spher2cart(parent,1,r2,theta2,phi2,x2,y2,z2)
      call distance_cart(x1,y1,z1,x2,y2,z2,d)

      return
      end


      subroutine distance_cart(x1,y1,z1,x2,y2,z2,d)
      implicit none
c
c     Purpose: to compute the distance between two points defined
c     by their cartesian coordinates
c
c     inputs
      double precision x1,y1,z1
      double precision x2,y2,z2
c     outputs
      double precision d

      d=dsqrt((x2-x1)**2.0D+0+(y2-y1)**2.0D+0+(z2-z1)**2.0D+0)

      return
      end


      subroutine spher2cart(parent,verif,r,theta,phi,x,y,z)
      implicit none
c     
c     Purpose: to convert (r,theta,phi) spherical coordinates
c     into (x,y,z) cartesian coordinates
c
c     Inputs:
c       + parent: name of the routine that is calling "spher2cart"
c       + verif: 1 if angles checking required; 0 otherwise
c       + r: radius (m), or altitude relative to the center of the planet
c       + theta: latitude (rad) [-pi/2,pi/2]
c       + phi: longitude rad) [0:2*pi]
c
c     Outputs:
c       + x (m)
c       + y (m)
c       + z (m)
c
      integer verif
      double precision r,theta,phi
      double precision x,y,z
      integer strlen
      character*100 parent,label
      label='subroutine spher2cart from: '//parent(1:strlen(parent))

      if (verif.eq.1) then
         call test_r(r,label)
         call test_theta(theta,label)
         call test_phi(phi,label)
      endif

      x=r*dcos(theta)*dcos(phi)
      y=r*dcos(theta)*dsin(phi)
      z=r*dsin(theta)
      
      return
      end



      subroutine cart2spher(x,y,z,r,theta,phi)
      implicit none
      include 'param.inc'
c     
c     Purpose: to convert (x,y,z) cartesian coordinates
c     into (r,theta,phi) spherical coordinates
c
c     Inputs:
c       + x (m)
c       + y (m)
c       + z (m)
c
c     Outputs:
c       + r: radius (m), or altitude relative to the center of the planet
c       + theta: latitude (rad) [-pi/2,pi/2]
c       + phi: longitude rad) [0:2*pi]
c
      double precision x,y,z
      double precision r,theta,phi
      double precision cosp,sinp,cost,sint
      double precision a,t1,t2

      a=dsqrt(x**2.0D+0+y**2.0D+0)
      r=dsqrt(x**2.0D+0+y**2.0D+0+z**2.0D+0)
      if (a.eq.0.0D+0) then
         cosp=1.0D+0
         sinp=0.0D+0
      else
         cosp=x/a
         sinp=y/a
      endif
      cost=a/r
      sint=z/r
      
      t1=dacos(dabs(cosp))
      if (x.ge.0.0D+0) then
         if (sinp.ge.0.0D+0) then
            phi=t1
         else
            phi=2.0D+0*pi-t1
         endif
      else
         if (sinp.ge.0.0D+0) then
            phi=pi-t1
         else
            phi=pi+t1
         endif
      endif

      t2=dacos(cost)
      if (sint.ge.0.0D+0) then
         theta=t2
      else
         theta=-t2
      endif
      
      return
      end
