c     Copyright (C) 2008-2009 Vincent Eymet
c
c     Planet_EMC is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     Planet_EMC is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with Planet_EMC; if not, see <http://www.gnu.org/licenses/>
c
      subroutine intersect_cell(dim,event,app,
     &     Pstart,u0,
     &     r1,r2,theta1,theta2,phi1,phi2,
     &     ca,cb,cc,
     &     intf,err_code,Pint,surface_out)
      implicit none
      include 'mpif.h'
      include 'max.inc'
c
c     Purpose: to find the intersection between a line defined
c     by a position and a direction, and the cell defined by
c     r1, r2, theta1, theta2, phi1, phi2 and
c     coefficients "a", "b" and "c" (equation: r=a*theta+b*phi+c).
c     This routine is used as a very last chance, when all other procedures
c     failed, but it costs a lot of CPU time.
c     This routine is based on "intersect_spherical_plane2" (it uses a dichotomy
c     procedure) extended for the general case of a cell.
c
c     Inputs:
c       + dim: dimension of vectors
c       + event: index of the current event
c       + app: if 1, real ground surface must be taken into account
c       + Pstart: (r,theta,phi) spherical coordinates of the line starting position
c       + u0: line propagation direction
c       + r1,r2,theta1,theta2,phi1,phi2: limits of the box
c       + ca, cb, cc: a, b and c coefficients for each quarter
c
c     Outputs:
c       + intf: 1 if an intersection was found; 0 otherwise
c       + err_code: error code when intf=0
c         - err_code=0: no intersection is possible with cell using those input arguments
c         - err_code=1: input position is out of cell
c         - err_code=2: an intersection was detected, but could not precisely be computed
c                       after Nit_max dichotomy iterations
c         - err_code=3: an intersection was successfully computed, but the ouput surface
c                       could not be determined
c       + Pint: (r,theta,phi) spherical coordinates of the intersection position, if any
c       + surface_out: index of the output surface
c

c     input
      integer dim,event,app
      double precision Pstart(1:Nmat_mx)
      double precision u0(1:Nmat_mx)
      double precision r1,r2,theta1,theta2,phi1,phi2
      double precision ca(1:4)
      double precision cb(1:4)
      double precision cc(1:4)
c     output
      integer intf,err_code,surface_out
      double precision Pint(1:Nmat_mx)
c     MPICH
      integer ierr,errorcode,np,pindex
      integer stat(MPI_STATUS_SIZE)
c     MPICH
c     temp
      double precision u(1:Nmat_mx)
      double precision n(1:Nmat_mx)
      integer ir,itheta,iphi,ons,ong
      integer in,cellf,surface
      integer i,gr,kmax,Nit_max,it
      double precision eps
      double precision P(1:Nmat_mx)
      double precision P1(1:Nmat_mx)
      double precision P2(1:Nmat_mx)
      double precision Pk(1:Nmat_mx),err
      double precision k1,k2
      double precision dmax
      double precision k
      character*100 log_file
      character*100 string
      character*100 stringp_file
      integer nl
      double precision pos(1:2),r_ground
      double precision epsilon_theta,epsilon_phi
      integer strlen
      character*100 label
c     MPICH
c     Get 'np', the number of processes
      call MPI_COMM_SIZE(MPI_COMM_WORLD,np,ierr)
c     Get 'pindex', the index of the current process
      call MPI_COMM_RANK(MPI_COMM_WORLD,pindex,ierr)
c     MPICH
      label='subroutine intersect_cell'
      epsilon_theta=1.0D-5 ! rad
      epsilon_phi=1.0D-5 ! rad
      Nit_max=1000
      eps=1.0D-5                ! m

      call logfile_name(pindex,event,log_file)
      string='./nlines'
      call stringp_file_name(pindex,string,stringp_file)

      intf=0
      err_code=0
      call copy_vector(dim,Pstart,P)
      call copy_vector(dim,u0,u)

      call get_nlines(log_file,stringp_file,nl)
      open(10,file=log_file(1:strlen(log_file)))
      do i=1,nl
         read(10,*)
      enddo
      if (nl.gt.0) then
         write(10,*)
      endif
      write(10,*) 'This is routine ',label(1:strlen(label))

c     get the order of magnitude of the maximal distance that can be travelled
c     among the current cell
      call distance_spher(label,r1,theta1,phi1,
     &     r2,theta2,phi2,dmax)
      kmax=int(2.0D+0*dmax)

c     if starting position is out of the input cell...
      call is_in_cell(P,app,
     &     r1,r2,
     &     theta1,theta2,
     &     phi1,phi2,
     &     ca,cb,cc,in)
      if (in.eq.0) then
         write(10,*) 'Input position is not in cell'
         close(10)
         err_code=1
         goto 666
c     ...then no solution is possible
      endif

c     otherwise, an intersection can be searched
      gr=0
      k1=0.0D+0
      do i=1,kmax
c     compute position "P+k*u0"
         k=dble(i)
         call jump_to(dim,P,u,k,Pk)
         call is_in_cell(Pk,app,
     &        r1,r2,
     &        theta1,theta2,
     &        phi1,phi2,
     &        ca,cb,cc,in)
         if (in.eq.0) then
            gr=1
            k2=k
            call copy_vector(dim,Pk,P2)
            goto 111
         else
            k1=k
            call copy_vector(dim,Pk,P1)
         endif
      enddo ! k
 111  continue
      if (gr.eq.0) then ! if no intersection could be found
         write(10,*) 'No intersection was detected'
         close(10)
         err_code=0
         goto 666
      else ! AH AH ! So there is an intersection with ground !
c     now dichotomy procedure between k1 and k2
         call distance_spher(label,P1(1),P1(2),P1(3),
     &        P2(1),P2(2),P2(3),err)
         it=0
         do while (err.gt.eps)
            it=it+1
            if (it.gt.Nit_max) then
               write(10,*) 'failure after max. number of iterations'
               close(10)
               err_code=2
               goto 666
            endif
            k=(k1+k2)/2.0D+0
            call jump_to(dim,P,u,k,Pk)
            call is_in_cell(Pk,app,
     &           r1,r2,
     &           theta1,theta2,
     &           phi1,phi2,
     &           ca,cb,cc,in)
            if (in.eq.0) then
               k2=k
               call copy_vector(dim,Pk,P2)
            else
               k1=k
               call copy_vector(dim,Pk,P1)
            endif
            call distance_spher(label,P1(1),P1(2),P1(3),
     &           P2(1),P2(2),P2(3),err)
         enddo
         call copy_vector(dim,Pk,Pint)
         write(10,*) 'Intersection was successfully computed:'
         write(10,*) 'Pint=',(Pint(i),i=1,dim)

c     Indentify output surface
         
c     test if position Pint is on the real ground (if possible)
         call is_on_ground(Pint,theta1,theta2,
     &        phi1,phi2,ca,cb,cc,ong)
         if (ong.eq.1) then
            ons=1
            surface_out=1
            goto 222
         else
            ons=0
         endif
c     if position Pint is not on the ground, it may be on the box surface
         if (ons.eq.0) then
            call is_on_surface(dim,r1,r2,
     &           theta1,theta2,
     &           phi1,phi2,
     &           Pint,ons,surface_out)
            if (ons.eq.1) then
               goto 222
            endif
         endif
c     finally, if everything else failed, it may be on a lateral surface, between ground and r1
         if (ons.eq.0) then
            if (app.eq.1) then
               pos(1)=Pint(2)
               pos(2)=Pint(3)
               call ground_radius(pos,theta1,theta2,
     &              phi1,phi2,ca,cb,cc,r_ground)
               if ((Pint(1).ge.r_ground-1.0D-4).and.
     &              (Pint(1).le.r2)) then
                  if (dabs(Pint(3)-phi1).lt.epsilon_phi) then
                     ons=1
                     surface_out=3
                     goto 222
                  else if (dabs(Pint(3)-phi2).lt.epsilon_phi) then
                     ons=1
                     surface_out=4
                     goto 222
                  else if (dabs(Pint(2)-theta1).lt.
     &                    epsilon_theta) then
                     ons=1
                     surface_out=5
                     goto 222
                  else if (dabs(Pint(2)-theta2).lt.epsilon_theta)
     &                    then
                     ons=1
                     surface_out=6
                     goto 222
                  endif
               endif ! Pint(1)>r_ground and Pint(1)<r2
            endif ! app=1
         endif ! ons=0
 222     continue

         write(10,*) 'ons=',ons
         if (ons.eq.0) then
            write(10,*) 'No output surface was identified'
            close(10)
            err_code=3
            goto 666
         else
            intf=1
            write(10,*) 'Output surface is:',surface_out
            close(10)
         endif ! ons
      endif
      
 666  continue
      return
      end



      subroutine intersect_spherical_plane2(dim,Pstart,u0,
     &     r1,r2,theta1,theta2,phi1,phi2,a,b,c,
     &     intf,err_code,Pint)
      implicit none
      include 'max.inc'
c
c     Purpose: to find the intersection (if any) between a line defined
c     by a position and a direction, and a "spherical plane" defined by
c     coefficients "a", "b" and "c" (equation: r=a*theta+b*phi+c).
c     This routine uses a dichotomy procedure, that is much more costly than
c     the standard Newton procedure (see the "intersect_spherical_plane" routine)
c     and is used only when the latter obviously failed.
c
c     Inputs:
c       + dim: dimension of vectors
c       + Pstart: (r,theta,phi) spherical coordinates of the line starting position
c       + u0: line propagation direction
c       + r1,r2,theta1,theta2,phi1,phi2: limits of the spherical box
c       + a, b and c: coefficients that define the "spherical plane"
c
c     Outputs:
c       + intf: 1 if an intersection was found; 0 otherwise
c       + err_code: error code when intf=0
c         - err_code=0: no intersection is possible with ground using those input arguments
c         - err_code=1: input position is below ground
c         - err_code=2: an intersection was detected, but could not precisely be computed
c                       after Nit_max dichotomy iterations
c       + Pint: (r,theta,phi) spherical coordinates of the intersection position, if any
c

c     input
      integer dim
      double precision Pstart(1:Nmat_mx)
      double precision u0(1:Nmat_mx)
      double precision r1,r2,theta1,theta2,phi1,phi2
      double precision a,b,c
c     output
      integer intf,err_code
      double precision Pint(1:Nmat_mx)
c     temp
      integer i,gr,kmax,Nit_max,it
      double precision eps
      double precision P(1:Nmat_mx)
      double precision Pk(1:Nmat_mx),err
      double precision k1,k2
      double precision dmax
      double precision pos(1:2),r_ground
      double precision k
      integer strlen
      character*100 label
      label='subroutine intersect_spherical_plane2'
      Nit_max=1000
      eps=1.0D-5                ! m

c     get the order of magnitude of the maximal distance that can be travelled
c     among the current cell
      call distance_spher(label,r1,theta1,phi1,r2,theta2,phi2,dmax)
      kmax=int(2.0D+0*dmax)

      intf=0
      err_code=0
      call copy_vector(dim,Pstart,P)

c     if starting position is below ground...
      pos(1)=P(2)
      pos(2)=P(3)
      call quarter_radius(pos,a,b,c,r_ground)
      if (P(1).lt.r_ground) then
         err_code=1
         goto 666
c     ...then no solution is possible
      endif

c     otherwise, an intersection can be searched
      gr=0
      k1=0.0D+0
      do i=1,kmax
c     compute position "P+k*u0"
         k=dble(i)
         call calc_err(dim,P,u0,k,a,b,c,Pk,err) ! output argument "err" is not used here
         pos(1)=Pk(2)
         pos(2)=Pk(3)
         call quarter_radius(pos,a,b,c,r_ground)
         if (Pk(1).le.r_ground) then
            gr=1
            k2=k
            goto 111
         else
            k1=k
         endif
      enddo ! k
 111  continue
      if (gr.eq.0) then ! if no intersection with ground could be found
         err_code=0
         goto 666
      else ! AH AH ! So there is an intersection with ground !
c     now dichotomy procedure between k1 and k2
         call calc_err(dim,P,u0,k1,a,b,c,Pk,err)
         it=0
         do while (err.gt.eps)
            it=it+1
            if (it.gt.Nit_max) then
               err_code=2
               goto 666
            endif
            k=(k1+k2)/2.0D+0
            call calc_err(dim,P,u0,k,a,b,c,Pk,err)
            pos(1)=Pk(2)
            pos(2)=Pk(3)
            call quarter_radius(pos,a,b,c,r_ground)
            if (Pk(1).le.r_ground) then
               k2=k
            else
               k1=k
            endif
         enddo
         intf=1
         call copy_vector(dim,Pk,Pint)
      endif
      
 666  continue
      return
      end




      subroutine intersect_spherical_plane(dim,Pstart,u0,
     &     r1,r2,theta1,theta2,phi1,phi2,a,b,c,
     &     intf,Pint)
      implicit none
      include 'max.inc'
c
c     Purpose: to find the intersection (if any) between a line defined
c     by a position and a direction, and a "spherical plane" defined by
c     coefficients "a", "b" and "c" (equation: r=a*theta+b*phi+c)
c     This routine uses the standard Newton method, that fails sometimes
c     (one trajectory each 50 millions)
c
c     Inputs:
c       + dim: dimension of vectors
c       + Pstart: (r,theta,phi) spherical coordinates of the line starting position
c       + u0: line propagation direction
c       + r1,r2,theta1,theta2,phi1,phi2: limits of the spherical box
c       + a, b and c: coefficients that define the "spherical plane"
c
c     Outputs:
c       + intf: 1 if an intersection was found; 0 otherwise
c       + Pint: (r,theta,phi) spherical coordinates of the intersection position, if any
c

c     input
      integer dim
      double precision Pstart(1:Nmat_mx)
      double precision u0(1:Nmat_mx)
      double precision r1,r2,theta1,theta2,phi1,phi2
      double precision a,b,c
c     output
      integer intf
      double precision Pint(1:Nmat_mx)
c     temp
      integer i
      integer intfs,iter
      integer tries,Ntries_max
      double precision P(1:Nmat_mx)
      double precision P0(1:Nmat_mx)
      double precision P1(1:Nmat_mx)
      double precision Pk(1:Nmat_mx)
      double precision pos(1:2),r_ground
      double precision k
      double precision err,derrdk
      double precision eps
      integer strlen
      character*100 label
c     Debug
      double precision nu0(1:Nmat_mx)
      double precision x0(1:Nmat_mx)
      double precision x(1:Nmat_mx)
      double precision seg(1:Nmat_mx)
      double precision err0,err1
c     Debug
      label='subroutine intersect_spherical_plane'
      Ntries_max=10

      intf=0
      tries=0
      call copy_vector(dim,Pstart,P)

c     first step: to find an intersection between the line and the sphere
c     defined by r=c. As we will always have a and b very small compared to c,
c     this is a good way to determine whether or not there will be an intersection
c     between the line and the "spherical plane"
c     call line_surface_r_intersect(0,dim,Pstart,u0,
c     &     c,theta1,theta2,phi1,phi2,intfs,P1)
 111  continue
c     take k=0 as starting position
      k=0.0D+0
      call calc_derrdk(dim,P,u0,k,a,b,c,derrdk)
      if (derrdk.eq.0.0D+0) then
         if (tries.le.Ntries_max) then
            tries=tries+1
            call calc_err(dim,P,u0,1.0D+1,a,b,c,Pk,err)
            call copy_vector(dim,Pk,P)
            goto 111
         else
            intf=0
            goto 666
         endif
      endif
      call calc_err(dim,P,u0,k,a,b,c,Pk,err)
      if (dabs(err/derrdk).gt.1.0D+20) then
         intf=0
         goto 666
      endif
      eps=1.0D-5                ! m
c     Debug
c     write(*,*) 'eps=',eps
c     Debug
      iter=0
      do while (dabs(err).gt.eps)
         iter=iter+1

         if (iter.gt.Niter_mx) then
c     sometimes it will simply not find any suitable solution, because there is none !
            intf=0
            goto 666
c     stop
         endif

         k=k-err/derrdk
         call calc_derrdk(dim,P,u0,k,a,b,c,derrdk)
         if (derrdk.eq.0.0D+0) then
            if (tries.le.Ntries_max) then
               tries=tries+1
               call calc_err(dim,P,u0,1.0D+1,a,b,c,Pk,err)
               call copy_vector(dim,Pk,P)
               goto 111
            else
               intf=0
               goto 666
            endif
         endif

         call calc_err(dim,P,u0,k,a,b,c,Pk,err)
         if (dabs(err).gt.1.0D+20) then
            intf=0
            goto 666
         endif
      enddo                     !  while err>eps
      intf=1
      call copy_vector(dim,Pk,Pint)
c     Debug
      pos(1)=Pint(2)
      pos(2)=Pint(3)
      call quarter_radius(pos,a,b,c,r_ground)
      if (dabs(Pint(1)-r_ground).gt.1.0D-2) then
         write(*,*) 'Error from ',
     &        label(1:strlen(label)),' :'
         write(*,*) 'Pstart=',(P(i),i=1,dim)
         write(*,*) 'Pint=',(Pint(i),i=1,dim)
         write(*,*) 'r_ground=',r_ground
         write(*,*) 'k=',k,' iter=',iter
         write(*,*) 'err=',err

         call normalize_vector(dim,u0,nu0)
         call spher2cart(label,1,P(1),P(2),P(3),
     &        x0(1),x0(2),x0(3)) ! x0: cartesian coordinates of "Pstart"
         call scalar_vector(dim,k,nu0,seg)

         write(*,*) 'seg=',(seg(i),i=1,dim)

         call add_vectors(dim,x0,seg,x)

         write(*,*) 'x0=',(x0(i),i=1,dim)
         write(*,*) 'x=',(x(i),i=1,dim)

         call cart2spher(x(1),x(2),x(3),P1(1),P1(2),P1(3))

         write(*,*) 'P1=',(P1(i),i=1,dim)

         pos(1)=P1(2)
         pos(2)=P1(3)
         call quarter_radius(pos,a,b,c,r_ground)
         err=dabs(P1(1)-r_ground)

         write(*,*) 'r_ground=',r_ground
         write(*,*) 'err=',err
         
         stop
      endif
c     Debug

 666  continue
      return
      end



      subroutine calc_derrdk(dim,Pstart,u0,k,a,b,c,derrdk)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute function "d(err)/d(k)" for routine "intersect_spherical_plane"
c
c     Inputs
c       + dim: dimension of vectors
c       + Pstart: (r,theta,phi) spherical coordinates of the starting position
c       + u0: line propagation direction
c       + k: position on the line
c       + a, b and c: coefficients that define the "spherical plane"
c     
c     Output:
c       + derrdk: function "d(err)/d(k)" at local position on line defined by "k"
c

c     input
      integer dim
      double precision Pstart(1:Nmat_mx)
      double precision u0(1:Nmat_mx)
      double precision k
      double precision a,b,c
c     output
      double precision derrdk
c     temp
      double precision k0,k1,dk
      double precision err0,err1
      double precision P0(1:Nmat_mx)
      double precision P1(1:Nmat_mx)

      dk=1.0D+0
      k0=k
      call calc_err(dim,Pstart,u0,k0,a,b,c,P0,err0)
      k1=k0+dk
      call calc_err(dim,Pstart,u0,k1,a,b,c,P1,err1)
      derrdk=(err1-err0)/dk

      return
      end




      subroutine calc_err(dim,Pstart,u0,k,a,b,c,Pk,err)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute function "err" for routine "intersect_spherical_plane"
c
c     Inputs
c       + dim: dimension of vectors
c       + Pstart: (r,theta,phi) spherical coordinates of the starting position
c       + u0: line propagation direction
c       + k: position on the line
c       + a, b and c: coefficients that define the "spherical plane"
c     
c     Output:
c       + Pk: (r,theta,phi) spherical coordinates of the position defined by "k"
c       + err: function "err", difference between the radius at position "k" on
c              the line, and the radius of the ground at (theta,phi) that is
c              known for position "k" on the line
c

c     input
      integer dim
      double precision Pstart(1:Nmat_mx)
      double precision u0(1:Nmat_mx)
      double precision k
      double precision a,b,c
c     output
      double precision err
c     temp
      double precision nu0(1:Nmat_mx)
      double precision x0(1:Nmat_mx)
      double precision seg(1:Nmat_mx)
      double precision x(1:Nmat_mx)
      double precision Pk(1:Nmat_mx)
      double precision pos(1:2)
      double precision r_ground
      integer strlen
      character*100 label
      label='subroutine calc_err'

      call normalize_vector(dim,u0,nu0)
      call spher2cart(label,1,Pstart(1),Pstart(2),Pstart(3),
     &     x0(1),x0(2),x0(3))   ! x0: cartesian coordinates of "Pstart"
      call scalar_vector(dim,k,nu0,seg)
      call add_vectors(dim,x0,seg,x)
      call cart2spher(x(1),x(2),x(3),Pk(1),Pk(2),Pk(3))
      pos(1)=Pk(2)
      pos(2)=Pk(3)
      call quarter_radius(pos,a,b,c,r_ground)
      err=dabs(Pk(1)-r_ground)

      return
      end




      subroutine dichotomy_planck(lambda1,lambda2,f1,f2,f0,eps,T,
     &     lambda0)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the value of "x0" that matches "f0"
c
c     Inputs:
c       + lambda1, lambda2: values of variable lambda
c       + f1, f2: values of the function at lambda1 and lambda2
c       + f0: value of the function that has to be inverted
c       + eps: relative error over f
c       + T: temperature [K]
c
c     Output:
c       + lambda0: value of variable lambda so that f(lambda0)=f0
c

c     inputs
      double precision lambda1,lambda2,f1,f2,f0,eps,T
c     output
      double precision lambda0
c     temp
      double precision x1,x2,x
      double precision F_wavlng
      double precision S,err
      double precision y1,y2,y
      integer strlen,it
      character*100 label
      label='subroutine dichotomy_planck'
      
      if (f0.le.0.0D+0) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Incorrect input parameter:'
         write(*,*) 'f0=',f0
         stop
      endif
      
      if (T.lt.0.0D+0) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Incorrect input parameter:'
         write(*,*) 'T=',T
         stop
      endif

      if (lambda1.eq.lambda2) then
         write(*,*) 'Error from ',label(1:strlen(label)),':'
         write(*,*) 'lambda1=',lambda1
         write(*,*) 'lambda2=',lambda2
         stop
      endif

      y1=f1-f0
      y2=f2-f0
      err=dabs(y2)/dabs(f0)
      x1=lambda1
      x2=lambda2

      if (f0.eq.f1) then
         lambda0=lambda1
         goto 123
      endif
      if (f0.eq.f2) then
         lambda0=lambda2
         goto 123
      endif

      it=0
      do while (err.gt.eps)
         it=it+1
         x=(x1+x2)/2.0D+0
         y=F_wavlng(x,T)-f0
         if (y*y1.gt.0.0D+0) then
            x1=x
            y1=y
         else
            x2=x
            y2=y
         endif
         err=dabs(y)/dabs(f0)
         if (it.gt.Niter_mx) then
            write(*,*) 'Error from ',label(1:strlen(label)),':'
            write(*,*) 'Too many iterations'
            write(*,*) 'x1=',x1
            write(*,*) 'x2=',x2
            write(*,*) 'f1=',f1
            write(*,*) 'f2=',f2
            write(*,*) 'f0=',f0
            stop
         endif
      enddo
      lambda0=x
      
c     numerical errors, only happen with pgf90 ...?
      if ((lambda0.lt.lambda1).or.(lambda0.gt.lambda2)) then
c     linear interpolation ?
         lambda0=lambda1+(lambda2-lambda1)/(f2-f1)*(f0-f1)
      endif

 123  continue
      if ((lambda0.lt.lambda1).or.(lambda0.gt.lambda2)) then
         write(*,*) 'Error from ',label(1:strlen(label)),':'
         write(*,*) 'lambda0 found=',lambda0,' is out of bounds:'
         write(*,*) 'lambda1=',lambda1
         write(*,*) 'lambda2=',lambda2
         write(*,*) 'f1=',f1
         write(*,*) 'f2=',f2
         write(*,*) 'f0=',f0
         stop
      endif

      return
      end
