c     Copyright (C) 2009 Vincent Eymet
c
c     "satellite" is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     This software is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with the present software; if not, see <http://www.gnu.org/licenses/>
c
      program satellite
      implicit none
      include 'mpif.h'
      include 'max.inc'
      include 'param.inc'
c     ------------------------------------------------------------------------
c     First, install mpich2:
c     retrieve the archive, unzip and untar it, install it on the system:
c     > ./configure --prefix=/home/[user]/parallele/mpich2
c     > make
c     > make install
c     
c     Before running MPD:
c     
c     1) create a ~/.mpd.conf file:
c     > echo secretword=[secretword] >> ~/.mpd.conf
c     > chmod 600 .mpd.conf
c     
c     2) allow your machine to connect by ssh to every other machine, without password:
c     create a dsa key:
c     > ssh-keygen -t dsa
c     then leave everything blank (only use 'enter' to proceed)
c     > cd .ssh
c     > cat id_dsa.pub >> authorized_keys
c     
c     3) create the mpd.hosts file: each line must contain the name of a machine
c     that is part of your cluster, by order of availability:
c     [host1].[domain]
c     [host2].[domain]
c     [host3].[domain]
c     etc
c     ------------------------------------------------------------------------
c     Running MPD:
c     > mpdboot -n [#]
c     with # the number of machines MPD has to run on
c     
c     use 'mpif77' in order to compile fortran 77 code in your makefile (or whatever)
c     
c     Code execution:
c     > mpirun -np [#] [executable]
c     with # the number of processes; in a master-slave algorithm, there is a master process and
c     'np-1' slave processes. If the master process is idle most of the time, it might be a good idea
c     to have 'np-1' equal to the number of physical cores, so that each (time consuming) process
c     uses a core.  In this case, 'np' is equal to the number of cores plus one. The total computation
c     time should be roughly divided by the number of cores (if communication times are negligeable,
c     and of course if the master process does not use too much time).
c     
c     Exit MPD:
c     > mpdallexit
c     ------------------------------------------------------------------------

c     MPICH
      integer ierr,errorcode,np,pindex
      integer stat(MPI_STATUS_SIZE)
      integer code,codep,proc,ret_code
      integer free(1:Nproc_mx),fpf,nsp
      integer nchunks,chunk
      integer ch_idx(1:Nproc_mx)
      integer end
c     MPICH
      integer dim
      double precision sat_theta,sat_phi,sat_z
      double precision point_theta,point_phi,op_alpha
      integer Nphi,Ntheta,Nb_gas
      double precision theta(0:Ntheta_mx)
      double precision phi(0:Nphi_mx)
      double precision Radius,rspace
      double precision lambda_gas_lo(1:Nb_gas_mx)
      double precision lambda_gas_hi(1:Nb_gas_mx)
      double precision solar_flux_toa(1:Ntheta_mx,1:Nphi_mx,
     &     1:Ntheta_mx,1:Nphi_mx,Nb_gas_mx) ! (W/m)
      double precision solar_inc_flux_toa(1:Ntheta_mx,1:Nphi_mx,
     &     1:Ntheta_mx,1:Nphi_mx,Nb_gas_mx) ! (W/m)
      double precision Psat(1:Nmat_mx)
      double precision Pground(1:Nmat_mx)
      integer Nint_sectors,i,j,sector,band
      integer int_sectors(1:Ntheta_mx*Nphi_mx,1:4)
      double precision Pc(1:Nb_gas_mx),Pc_tot
      double precision inc_Pc(1:Nb_gas_mx),inc_Pc_tot
c     partial results
      double precision Pc_1s(1:Nb_gas_mx),Pc_tot_1s
      double precision inc_Pc_1s(1:Nb_gas_mx),inc_Pc_tot_1s
c     temp
      integer strlen
      character*100 label
      label='program satellite'

c     MPICH
      call MPI_INIT(ierr)
      if (ierr.NE.MPI_SUCCESS) THEN
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'MPI could not be initialized'
         call MPI_ABORT(MPI_COMM_WORLD,errorcode,ierr)
      endif
c     Get 'np', the number of processes
      call MPI_COMM_SIZE(MPI_COMM_WORLD,np,ierr)
      if (np.lt.2) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Number of processes is:',np
         write(*,*) 'and it should be greater than 1'
         stop
      endif
      if (np.gt.Nproc_mx) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Number of processes is:',np
         write(*,*) 'is greater than Nproc_mx=',Nproc_mx
         stop
      endif
c     Get 'pindex', the index of the current process
      call MPI_COMM_RANK(MPI_COMM_WORLD,pindex,ierr)
c     MPICH

c     MPICH
      if (pindex.eq.0) then
c     MPICH
         write(*,*) 'Initializing...'
         call init(dim,Pc,Pc_tot)
c     read data
         call read_data(sat_theta,sat_phi,sat_z,
     &        point_theta,point_phi,op_alpha)
c     read TOA fluxes and uncertainties
         call read_Ftoa(Ntheta,Nphi,Nb_gas,theta,phi,
     &        lambda_gas_lo,lambda_gas_hi,
     &        Radius,rspace,
     &        solar_flux_toa,
     &        solar_inc_flux_toa)
c     compute Psat and Pground
         call set_positions(sat_z,Radius,
     &        sat_theta,sat_phi,point_theta,point_phi,
     &        Psat,Pground)
c     look for errors
         write(*,*) 'Looking for known errors...'
         call look_for_errors(dim,Psat,Pground,op_alpha,rspace)
         write(*,*) '...none found, resuming'
c     compute sectors that emit towards the satellite
         call intersected_sectors(dim,Ntheta,Nphi,theta,phi,
     &        Radius,rspace,Psat,Pground,op_alpha,
     &        Nint_sectors,int_sectors)
c     set Nchunks
         call set_Nchunks(Nint_sectors,Nchunks)
c     sending data to all processes
c     Debug
         write(*,*) 'Sending data to all processes...'
c     Debug
         code=1
         do proc=1,np-1
            call MPI_SEND(code,1,MPI_INTEGER
     &           ,proc,1,MPI_COMM_WORLD,ierr)
         enddo
         call send_global_data(dim,
     &        Ntheta,Nphi,Nb_gas,Nint_sectors,
     &        int_sectors,Radius,rspace,op_alpha,theta,phi,
     &        Psat,Pground,solar_flux_toa,solar_inc_flux_toa)
c     Debug
         write(*,*) '...done'
c     Debug
         
         do proc=1,np-1
            free(proc)=1
         enddo
         end=0
         chunk=0
c     Debug
         write(*,*) 'Now computing...'
c     Debug

         do while (end.eq.0)
 444        continue
            call find_free_process(free,np-1,fpf,proc)
            if (fpf.eq.1) then  ! a free child process found
               chunk=chunk+1
               if (chunk.le.Nchunks) then
                  ch_idx(proc)=chunk
                  code=2 ! means a computation must be held
                  call MPI_SEND(code,1,MPI_INTEGER
     &                 ,proc,1,MPI_COMM_WORLD,ierr)
                  free(proc)=0  ! child process index 'proc' is busy
               else             ! chunk>nchunks
                  free(proc)=2  ! means child process index 'proc' has been stopped
               endif
               goto 444
            else                ! no free child process found
               call stopped_processes(free,np-1,nsp)
               if (nsp.ge.np-1) then
                  end=1
                  goto 111
               else
c     waiting for news from slave processes...
c     reception of a slave process index: 'proc'
                  call MPI_RECV(proc,1,MPI_INTEGER
     &                 ,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,
     &                 stat,ierr)
c     reception of a 'ret_code' from process index 'proc'
                  call MPI_RECV(ret_code,1,MPI_INTEGER
     &                 ,proc,MPI_ANY_TAG,MPI_COMM_WORLD,
     &                 stat,ierr)

                  if (ret_code.eq.0) then
c     process index 'proc' requires data to compute something
                     call MPI_SEND(ch_idx(proc),1,MPI_INTEGER
     &                    ,proc,1,MPI_COMM_WORLD,ierr)
                  else if (ret_code.eq.1) then
c     reception of partial computation results from process index 'proc'
                     call receive_partial_results(proc,Pc_1s,Pc_tot_1s,
     &                    inc_Pc_1s,inc_Pc_tot_1s)
                     free(proc)=1 ! child process index 'proc' is free again
c     update total results
                     call update_results(Nb_gas,
     &                    Pc_1s,inc_Pc_1s,Pc_tot_1s,inc_Pc_tot_1s,
     &                    Pc,inc_Pc,Pc_tot,inc_Pc_tot)

                  else
                     write(*,*) 'Error from ',
     &                    label(1:strlen(label)),' :'
                     write(*,*) 'ret_code=',ret_code
                     stop
                  endif

               endif            ! nsp=np-1
            endif               ! fpf=0 or 1
 111        continue
         enddo                  ! while end=0

c     stopping all child processes
         do proc=1,np-1
            code=0
            call MPI_SEND(code,1,MPI_INTEGER,proc,1,MPI_COMM_WORLD,ierr)
         enddo
c     Debug
         write(*,*) '...done'
         write(*,*) 'Pc(band 1)=',Pc(1),'+/-',inc_Pc(1),' W/m'
         write(*,*) 'Pc_tot=',Pc_tot,'+/-',inc_Pc_tot,' W/m'
c     Debug
c     writing results to file
         call write_results(Nb_gas,
     &        lambda_gas_lo,lambda_gas_hi,
     &        Pc,Pc_tot,inc_Pc,inc_Pc_tot)
      
      else ! every other process

 333     continue
c     wait for "codep" from root process
         call MPI_RECV(codep,1,MPI_INTEGER
     &        ,0,MPI_ANY_TAG,MPI_COMM_WORLD,stat,ierr)

c     codep=0 means exit
         if (codep.eq.0) then
            goto 666
c     codep=1 means data reception for solar computation
         else if (codep.eq.1) then
            call receive_global_data(dim,
     &           Ntheta,Nphi,Nb_gas,Nint_sectors,
     &           int_sectors,Radius,rspace,op_alpha,theta,phi,
     &           Psat,Pground,solar_flux_toa,solar_inc_flux_toa)

c     codep=2 means the process has to run a computation
         else if (codep.eq.2) then
c     sending process index to process 0
            call MPI_SEND(pindex,1,MPI_INTEGER,0,1,MPI_COMM_WORLD,ierr)
c     sending integer '0' to process 0, meaning a sector index is required
            call MPI_SEND(0,1,MPI_INTEGER,0,1,MPI_COMM_WORLD,ierr)
c     receiving sector index from master process
            call MPI_RECV(sector,1,MPI_INTEGER
     &           ,0,MPI_ANY_TAG,MPI_COMM_WORLD,
     &           stat,ierr)
c     computation for provided sector index
            call solid_angles(dim,Ntheta,Nphi,theta,phi,Nb_gas,
     &           Psat,Pground,
     &           rspace,op_alpha,
     &           Nint_sectors,int_sectors,sector,
     &           solar_flux_toa,solar_inc_flux_toa,
     &           Pc_1s,Pc_tot_1s,inc_Pc_1s,inc_Pc_tot_1s)
c     sending process index 'pindex' to master process
            call MPI_SEND(pindex,1,MPI_INTEGER,0,1,MPI_COMM_WORLD,ierr)
c     sending integer '1' to process 0, meaning partial results will be send now
            call MPI_SEND(1,1,MPI_INTEGER,0,1,MPI_COMM_WORLD,ierr)
c     sendind partial results to the process 0
            call send_partial_results(Pc_1s,Pc_tot_1s,
     &           inc_Pc_1s,inc_Pc_tot_1s)

         else ! unknown code
            write(*,*) 'Error from ',
     &           label(1:strlen(label)),' :'
            write(*,*) 'Unknown codep=',codep
            write(*,*) 'received by process:',pindex
            stop
         endif                  ! codep

c     wait for a new execution code
         goto 333
 666     continue
c     MPICH
      endif                     ! pindex
c     MPICH

c     MPICH
      call MPI_FINALIZE(ierr)
c     MPICH

      end
