c     Copyright (C) 2008-2009 Vincent Eymet
c
c     Planet_EMC is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     Planet_EMC is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with Planet_EMC; if not, see <http://www.gnu.org/licenses/>
c
      subroutine is_on_edge(P,r1,r2,theta1,theta2,phi1,phi2,
     &     app,ca,cb,cc,ioe,edge)
      implicit none
      include 'max.inc'
c
c     Purpose: to check whether or not the provided position is on an edge of the spherical cell
c
c     Inputs:
c       + P: (r,theta,phi) spherical coordinates of the position to test
c       + r1,r2,theta1,theta2,phi1,phi2: limits of the cell
c       + app: if 1, real ground surface must be computed for surface 1
c       + ca,cb,cc: coefficients for computing ground radius
c
c     Output:
c       + ioe: 1: particle is near edge; 0 otherwise
c       + edge: index of the edge the particle is on, if any found
c
c     input
      double precision P(1:Nmat_mx)
      double precision r1,r2,theta1,theta2,phi1,phi2
      integer app
      double precision ca(1:4)
      double precision cb(1:4)
      double precision cc(1:4)
c     output
      integer ioe,edge
c     temp
      double precision pos(1:2)
      double precision r_ground
      double precision dist_min
      double precision dr,dtheta,dphi
      integer strlen
      character*100 label
      label='subroutine is_on_edge'
      dr=1.0D-6 ! m
      dtheta=1.0D-5 ! rad
      dphi=1.0D-5 ! rad
      
      ioe=0
c     testing edges 1-4
      if (app.eq.1) then
         pos(1)=P(2)
         pos(2)=P(3)
         call ground_radius(pos,theta1,theta2,phi1,phi2,
     &     ca,cb,cc,r_ground)
         if ((P(2).eq.theta1).and.(P(1).eq.r_ground)) then
            ioe=1
            edge=1
         else if ((P(3).eq.phi1).and.(P(1).eq.r_ground)) then
            ioe=1
            edge=2
         else if ((P(2).eq.theta2).and.(P(1).eq.r_ground)) then
            ioe=1
            edge=3
         else if ((P(3).eq.phi2).and.(P(1).eq.r_ground)) then
            ioe=1
            edge=4
         endif
c         if ((dabs(P(2)-theta1).lt.dtheta).and.
c     &        (dabs(P(1)-r_ground).lt.dr)) then
c            ioe=1
c            edge=1
c         else if ((dabs(P(3)-phi1).lt.dphi).and.
c     &           (dabs(P(1)-r_ground).lt.dr)) then
c            ioe=1
c            edge=2
c         else if ((dabs(P(2)-theta2).lt.dtheta).and.
c     &           (dabs(P(1)-r_ground).lt.dr)) then
c            ioe=1
c            edge=3
c         else if ((dabs(P(3)-phi2).lt.dphi).and.
c     &           (dabs(P(1)-r_ground).lt.dr)) then
c            ioe=1
c            edge=4
c         endif
      else
        if ((P(2).eq.theta1).and.(P(1).eq.r1)) then
           ioe=1
           edge=1
        else if ((P(3).eq.phi1).and.(P(1).eq.r1)) then
           ioe=1
           edge=2
        else if ((P(2).eq.theta2).and.(P(1).eq.r1)) then
           ioe=1
           edge=3
        else if ((P(3).eq.phi2).and.(P(1).eq.r1)) then
           ioe=1
           edge=4
        endif
c         if ((dabs(P(2)-theta1).lt.dtheta).and.
c     &        (dabs(P(1)-r1).lt.dr)) then
c            ioe=1
c            edge=1
c         else if ((dabs(P(3)-phi1).lt.dphi).and.
c     &           (dabs(P(1)-r1).lt.dr)) then
c            ioe=1
c            edge=2
c         else if ((dabs(P(2)-theta2).lt.dtheta).and.
c     &           (dabs(P(1)-r1).lt.dr)) then
c            ioe=1
c            edge=3
c         else if ((dabs(P(3)-phi2).lt.dphi).and.
c     &           (dabs(P(1)-r1).lt.dr)) then
c            ioe=1
c            edge=4
c         endif
      endif
c     testing edges 5-8
      if ((P(2).eq.theta1).and.(P(1).eq.r2)) then
         ioe=1
         edge=5
      else if ((P(3).eq.phi1).and.(P(1).eq.r2)) then
         ioe=1
         edge=6
      else if ((P(2).eq.theta2).and.(P(1).eq.r2)) then
         ioe=1
         edge=7
      else if ((P(3).eq.phi2).and.(P(1).eq.r2)) then
         ioe=1
         edge=8
      endif
c      if ((dabs(P(2)-theta1).lt.dtheta).and.
c     &     (dabs(P(1)-r2).lt.dr)) then
c         ioe=1
c         edge=5
c      else if ((dabs(P(3)-phi1).lt.dphi).and.
c     &        (dabs(P(1)-r2).lt.dr)) then
c         ioe=1
c         edge=6
c      else if ((dabs(P(2)-theta2).lt.dtheta).and.
c     &        (dabs(P(1)-r2).lt.dr)) then
c         ioe=1
c         edge=7
c      else if ((dabs(P(3)-phi2).lt.dphi).and.
c     &        (dabs(P(1)-r2).lt.dr)) then
c         ioe=1
c         edge=8
c      endif

c     testing edges 9-12
      if ((P(2).eq.theta1).and.(P(3).eq.phi1)) then
         ioe=1
         edge=9
      else if ((P(2).eq.theta1).and.(P(3).eq.phi2)) then
         ioe=1
         edge=10
      else if ((P(2).eq.theta2).and.(P(3).eq.phi1)) then
         ioe=1
         edge=11
      else if ((P(2).eq.theta2).and.(P(3).eq.phi2)) then
         ioe=1
         edge=12
      endif
         

c     Debug
c     if (ioe.eq.1) then
c     write(*,*) label(1:strlen(label)),' : edge detected'
c      endif
c     Debug

      return
      end




      subroutine is_near_corner(P,r1,r2,theta1,theta2,phi1,phi2,
     &     app,ca,cb,cc,inc,corner)
      implicit none
      include 'max.inc'
c
c     Purpose: to check whether or not the provided position is near a corner of the spherical cell
c
c     Inputs:
c       + P: (r,theta,phi) spherical coordinates of the position to test
c       + r1,r2,theta1,theta2,phi1,phi2: limits of the cell
c       + app: if 1, real ground surface must be computed for surface 1
c       + ca,cb,cc: coefficients for computing ground radius
c
c     Output:
c       + inc: 1: particle is near corner; 0 otherwise
c       + corner: index of the corner the particle is near from, if any found
c
c     input
      double precision P(1:Nmat_mx)
      double precision r1,r2,theta1,theta2,phi1,phi2
      integer app
      double precision ca(1:4)
      double precision cb(1:4)
      double precision cc(1:4)
c     output
      integer inc,corner
c     temp
      double precision pos(1:2)
      double precision Pc1(1:Nmat_mx)
      double precision Pc2(1:Nmat_mx)
      double precision Pc3(1:Nmat_mx)
      double precision Pc4(1:Nmat_mx)
      double precision Pc5(1:Nmat_mx)
      double precision Pc6(1:Nmat_mx)
      double precision Pc7(1:Nmat_mx)
      double precision Pc8(1:Nmat_mx)
      double precision dist2corner(1:8)
      double precision dist_min
      integer i,ncf,corner_idx(1:8)
      integer strlen
      character*100 label
      label='subroutine is_near_corner'
      dist_min=1.0D-8 ! m

c     coordinates of corners 1-4
      Pc1(2)=theta1
      Pc1(3)=phi2
      Pc2(2)=theta1
      Pc2(3)=phi1
      Pc3(2)=theta2
      Pc3(3)=phi1
      Pc4(2)=theta2
      Pc4(3)=phi2
      if (app.eq.1) then
         pos(1)=Pc1(2)
         pos(2)=Pc1(3)
         call ground_radius(pos,theta1,theta2,phi1,phi2,
     &     ca,cb,cc,Pc1(1))
         pos(1)=Pc2(2)
         pos(2)=Pc2(3)
         call ground_radius(pos,theta1,theta2,phi1,phi2,
     &     ca,cb,cc,Pc2(1))
         pos(1)=Pc3(2)
         pos(2)=Pc3(3)
         call ground_radius(pos,theta1,theta2,phi1,phi2,
     &     ca,cb,cc,Pc3(1))
         pos(1)=Pc4(2)
         pos(2)=Pc4(3)
         call ground_radius(pos,theta1,theta2,phi1,phi2,
     &     ca,cb,cc,Pc4(1))
      else
         Pc1(1)=r1
         Pc2(1)=r1
         Pc3(1)=r1
         Pc4(1)=r1
      endif
c     coordinates of corners 5-8
      Pc5(1)=r2
      Pc5(2)=theta1
      Pc5(3)=phi2
      Pc6(1)=r2
      Pc6(2)=theta1
      Pc6(3)=phi1
      Pc7(1)=r2
      Pc7(2)=theta2
      Pc7(3)=phi1
      Pc8(1)=r2
      Pc8(2)=theta2
      Pc8(3)=phi2
c     Debug
      write(*,*) 'Pc1=',(Pc1(i),i=1,3)
      write(*,*) 'Pc2=',(Pc2(i),i=1,3)
      write(*,*) 'Pc3=',(Pc3(i),i=1,3)
      write(*,*) 'Pc4=',(Pc4(i),i=1,3)
      write(*,*) 'Pc5=',(Pc5(i),i=1,3)
      write(*,*) 'Pc6=',(Pc6(i),i=1,3)
      write(*,*) 'Pc7=',(Pc7(i),i=1,3)
      write(*,*) 'Pc8=',(Pc8(i),i=1,3)
c     Debug
c     computing distance to each corner
      call distance_spher(label,P(1),P(2),P(3),Pc1(1),Pc1(2),Pc1(3),
     &     dist2corner(1))
      call distance_spher(label,P(1),P(2),P(3),Pc2(1),Pc2(2),Pc2(3),
     &     dist2corner(2))
      call distance_spher(label,P(1),P(2),P(3),Pc3(1),Pc3(2),Pc3(3),
     &     dist2corner(3))
      call distance_spher(label,P(1),P(2),P(3),Pc4(1),Pc4(2),Pc4(3),
     &     dist2corner(4))
      call distance_spher(label,P(1),P(2),P(3),Pc5(1),Pc5(2),Pc5(3),
     &     dist2corner(5))
      call distance_spher(label,P(1),P(2),P(3),Pc6(1),Pc6(2),Pc6(3),
     &     dist2corner(6))
      call distance_spher(label,P(1),P(2),P(3),Pc7(1),Pc7(2),Pc7(3),
     &     dist2corner(7))
      call distance_spher(label,P(1),P(2),P(3),Pc8(1),Pc8(2),Pc8(3),
     &     dist2corner(8))
c     Debug
      do i=1,8
         write(*,*) 'dist2corner(',i,')=',dist2corner(i)
      enddo
c     Debug
c     computing the number of corners the particle is near from
      ncf=0
      do i=1,8
         if (dabs(dist2corner(i)).le.dist_min) then
            ncf=ncf+1
            corner_idx(ncf)=i
         endif
      enddo ! i
c     there should be only one !
      if (ncf.gt.1) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Number of corners found:',ncf
         do i=1,ncf
            write(*,*) 'corner index:',corner_idx(i)
         enddo ! i
      endif
      if (ncf.eq.1) then
         inc=1
         corner=corner_idx(1)
      else
         inc=0
      endif

      return
      end




      subroutine is_in_cell(coord,app,r1,r2,theta1,theta2,
     &     phi1,phi2,ca,cb,cc,in)
      implicit none
      include 'max.inc'
      include 'param.inc'
c
c     Purpose: to identify whether or not the provided point is within
c     the provided cell
c
c     Inputs:
c       + coord: (r,theta,phi) spherical coordinates of the point to test
c       + app: if 1, the real ground surface must be computed
c       + r1, theta1, phi1, r2, theta2, phi2: limits of the spherical box
c       + ca,cb,cc: a, b and c coefficients for the 4 quarters of the angular sector
c
c     Output:
c       + in:
c         - in=0: point 'coord' IS NOT within the cell
c         - in=1: point 'coord' IS within the cell
c

c     input
      double precision coord(1:Nmat_mx)
      integer app
      double precision r1,r2,theta1,theta2,phi1,phi2
c     output
      integer in
c     temp
      integer inc
      double precision pos(1:2),r_ground
      double precision ca(1:4)
      double precision cb(1:4)
      double precision cc(1:4)
      double precision dr
      integer strlen,i
      character*100 label
      label='subroutine is_in_cell'
      dr=1.0D-4

      in=0
      if (app.eq.0) then
         call is_in_box(coord,r1,r2,theta1,theta2,
     &        phi1,phi2,in)
      else
         call is_in_column(coord,theta1,theta2,phi1,phi2,inc)
         pos(1)=coord(2)
         pos(2)=coord(3)
         call ground_radius(pos,theta1,theta2,phi1,phi2,
     &        ca,cb,cc,r_ground)
         if ((inc.eq.1).and.
     &        (coord(1).ge.r_ground-dr).and.(coord(1).le.r2)) then
            in=1
         endif
      endif
      
      return
      end




      subroutine is_in_box(coord,r1,r2,theta1,theta2,
     &     phi1,phi2,in)
      implicit none
      include 'max.inc'
      include 'param.inc'
c
c     Purpose: to identify whether or not the provided point is within
c     the provided spherical box
c
c     Inputs:
c       + coord: (r,theta,phi) spherical coordinates of the point to test
c       + r1, theta1, phi1, r2, theta2, phi2: limits of the spherical box
c
c     Output:
c       + in:
c         - in=0: point 'coord' IS NOT within the box
c         - in=1: point 'coord' IS within the box
c

c     input
      double precision coord(1:Nmat_mx)
      double precision r1,r2,theta1,theta2,phi1,phi2
c     output
      integer in
c     temp
      double precision epsilon_r
      double precision epsilon_theta
      double precision epsilon_phi
      integer strlen,i
      character*100 label
      parameter(epsilon_theta=1.0D-5)
      parameter(epsilon_phi=1.0D-5)
      epsilon_r=1.0D-11*r1
      label='subroutine is_in_box'

      call test_box(r1,r2,theta1,theta2,phi1,phi2,label)
      if ((coord(1).ge.r1-epsilon_r).and.
     &     (coord(1).le.r2+epsilon_r)
     &     .and.(coord(2).ge.theta1-epsilon_theta).and.
     &     (coord(2).le.theta2+epsilon_theta)) then
         if (((coord(3).ge.phi1-epsilon_phi).and.
     &     (coord(3).le.phi2+epsilon_phi)).or.
     &        ((dabs(coord(3)).lt.epsilon_phi).and.
     &        (dabs(phi2-2.0D+0*pi).lt.epsilon_phi))) then
            in=1
         else
            in=0
c     Debug
c            write(*,*) 'This is routine ',label(1:strlen(label))
c            write(*,*) 'r1=',r1
c            write(*,*) 'r2=',r2
c            write(*,*) 'theta1=',theta1
c            write(*,*) 'theta2=',theta2
c            write(*,*) 'phi1=',phi1
c            write(*,*) 'phi2=',phi2
c            write(*,*) 'coord=',(coord(i),i=1,3)
c     Debug
         endif
      else
         in=0
      endif
      
      return
      end


      double precision function volume_box(r1,r2,theta1,theta2,
     &     phi1,phi2)
      implicit none
      double precision r1,r2,theta1,theta2,phi1,phi2
      character*100 label
c     
c     Purpose: to compute the volume of a given sphere portion
c     defined by radiuses r1 and r2, latitude angles theta1 and theta2,
c     longitude angles phi1 and phi2
c
c     Inputs:
c       + r1: first radius (m)
c       + r2: second radius (m), must be greater or equal to r1
c       + theta1: first latitude angle (rad)
c       + theta2: second latitude angle (rad), must be greater or equal to theta1
c       + phi1: first longitude angle (rad)
c       + phi2: second longitude angle (rad), must be greater or equal to phi1
c
c     Output:
c       + volume_box: volume (m)
c     
      label='subroutine volume_box'
      call test_box(r1,r2,theta1,theta2,phi1,phi2,label)
      
      volume_box=(phi2-phi1)
     &     *(dsin(theta2)-dsin(theta1))
     &     *(r2**3.0D+0-r1**3.0D+0)/3.0D+0
      
      return
      end
